@@ -1,10 +1,10 @@
 #!/usr/bin/python3
 
-import codecs
-import io
+import hashlib
 import json
 import os.path
 import re
+import ssl
 import subprocess
 import sys
 
@@ -32,7 +32,6 @@ class Gencontrol(Base):
 
         self.abiname = config_entry['abiname']
         self.vars = {
-            'template': 'linux-image-%s-signed-template' % arch,
             'upstreamversion': self.version.linux_upstream,
             'version': self.version.linux_version,
             'source_basename': re.sub(r'-[\d.]+$', '',
@@ -45,6 +44,8 @@ class Gencontrol(Base):
         }
         self.vars['source_suffix'] = \
             self.changelog[0].source[len(self.vars['source_basename']):]
+        self.vars['template'] = \
+            'linux-image%(source_suffix)s-%(arch)s-signed-template' % self.vars
 
         self.package_dir = 'debian/%(template)s' % self.vars
         self.template_top_dir = (self.package_dir
@@ -83,12 +84,9 @@ class Gencontrol(Base):
             kw_proc = subprocess.Popen(
                 ['kernel-wedge', 'gen-control', vars['abiname']],
                 stdout=subprocess.PIPE,
+                text=True,
                 env=kw_env)
-            if not isinstance(kw_proc.stdout, io.IOBase):
-                udeb_packages = read_control(io.open(kw_proc.stdout.fileno(),
-                                                     closefd=False))
-            else:
-                udeb_packages = read_control(io.TextIOWrapper(kw_proc.stdout))
+            udeb_packages = read_control(kw_proc.stdout)
             kw_proc.wait()
             if kw_proc.returncode != 0:
                 raise RuntimeError('kernel-wedge exited with code %d' %
@@ -117,7 +115,12 @@ class Gencontrol(Base):
     def do_arch_setup(self, vars, makeflags, arch, extra):
         super(Gencontrol, self).do_main_setup(vars, makeflags, extra)
 
-        makeflags['ABINAME'] = vars['abiname']
+        if self.version.linux_modifier is None:
+            abiname_part = '-%s' % self.config.merge('abi', arch)['abiname']
+        else:
+            abiname_part = ''
+        makeflags['ABINAME'] = vars['abiname'] = \
+            self.config['version', ]['abiname_base'] + abiname_part
 
     def do_arch_packages(self, packages, makefile, arch, vars, makeflags,
                          extra):
@@ -171,8 +174,8 @@ class Gencontrol(Base):
 
     def do_flavour_packages(self, packages, makefile, arch, featureset,
                             flavour, vars, makeflags, extra):
-        if not (self.config.merge('build', arch, featureset, flavour)
-                .get('signed-code', False)):
+        config_build = self.config.merge('build', arch, featureset, flavour)
+        if not config_build.get('signed-code', False):
             return
 
         # In a quick build, only build the quick flavour (if any).
@@ -184,7 +187,16 @@ class Gencontrol(Base):
         image_suffix = '%(abiname)s%(localversion)s' % vars
         image_package_name = 'linux-image-%s-unsigned' % image_suffix
 
-        self.image_packages.append((image_suffix, image_package_name))
+        # Verify that this flavour is configured to support Secure Boot,
+        # and get the trusted certificates filename.
+        with open('debian/%s/boot/config-%s' %
+                  (image_package_name, image_suffix)) as f:
+            kconfig = f.readlines()
+        assert 'CONFIG_EFI_STUB=y\n' in kconfig
+        assert 'CONFIG_LOCK_DOWN_IN_EFI_SECURE_BOOT=y\n' in kconfig
+        cert_file_name = config_build['trusted-certs']
+        self.image_packages.append((image_suffix, image_package_name,
+                                    cert_file_name))
 
         packages['source']['Build-Depends'].append(
             image_package_name
@@ -264,17 +276,18 @@ class Gencontrol(Base):
             re.sub(r'\+b(\d+)$', r'.b\1',
                    re.sub(r'-', r'+', vars['imagebinaryversion']))
 
-        with codecs.open(self.template_debian_dir + '/changelog', 'w',
-                         'utf-8') as f:
+        with open(self.template_debian_dir + '/changelog', 'w',
+                  encoding='utf-8') as f:
             f.write(self.substitute('''\
-linux-signed-@arch@ (@signedsourceversion@) @distribution@; urgency=@urgency@
+linux-signed@source_suffix@-@arch@ (@signedsourceversion@) @distribution@; urgency=@urgency@
 
   * Sign kernel from @source@ @imagebinaryversion@
 
 ''',
                                     vars))
 
-            with codecs.open('debian/changelog', 'r', 'utf-8') as changelog_in:
+            with open('debian/changelog', 'r', encoding='utf-8') \
+                 as changelog_in:
                 # Ignore first two header lines
                 changelog_in.readline()
                 changelog_in.readline()
@@ -283,18 +296,67 @@ linux-signed-@arch@ (@signedsourceversion@) @distribution@; urgency=@urgency@
                     f.write(d)
 
     def write_files_json(self):
+        # Can't raise from a lambda function :-(
+        def raise_func(e):
+            raise e
+
+        # Some functions in openssl work with multiple concatenated
+        # PEM-format certificates, but others do not.
+        def get_certs(file_name):
+            certs = []
+            BEGIN, MIDDLE = 0, 1
+            state = BEGIN
+            with open(file_name) as f:
+                for line in f:
+                    if line == '-----BEGIN CERTIFICATE-----\n':
+                        assert state == BEGIN
+                        certs.append([])
+                        state = MIDDLE
+                    elif line == '-----END CERTIFICATE-----\n':
+                        assert state == MIDDLE
+                        state = BEGIN
+                    else:
+                        assert line[0] != '-'
+                        assert state == MIDDLE
+                    certs[-1].append(line)
+            assert state == BEGIN
+            return [''.join(cert_lines) for cert_lines in certs]
+
+        def get_cert_fingerprint(cert, algo):
+            hasher = hashlib.new(algo)
+            hasher.update(ssl.PEM_cert_to_DER_cert(cert))
+            return hasher.hexdigest()
+
         all_files = {'packages': {}}
 
-        for image_suffix, image_package_name in self.image_packages:
+        for image_suffix, image_package_name, cert_file_name in \
+                self.image_packages:
+            package_dir = 'debian/%s' % image_package_name
             package_files = []
+            package_modules = []
             package_files.append({'sig_type': 'efi',
                                   'file': 'boot/vmlinuz-%s' % image_suffix})
+            for root, dirs, files in os.walk('%s/lib/modules' % package_dir,
+                                             onerror=raise_func):
+                for name in files:
+                    if name.endswith('.ko'):
+                        package_modules.append(
+                            '%s/%s' %
+                            (root[(len(package_dir) + 1):], name))
+            package_modules.sort()
+            for module in package_modules:
+                package_files.append(
+                    {'sig_type': 'linux-module',
+                     'file': module})
+            package_certs = [get_cert_fingerprint(cert, 'sha256')
+                             for cert in get_certs(cert_file_name)]
+            assert len(package_certs) >= 1
             all_files['packages'][image_package_name] = {
-                'trusted_certs': [],
+                'trusted_certs': package_certs,
                 'files': package_files
             }
 
-        with codecs.open(self.template_top_dir + '/files.json', 'w') as f:
+        with open(self.template_top_dir + '/files.json', 'w') as f:
             json.dump(all_files, f)
 
     def write_source_lintian_overrides(self):
